#define CORE_NAME "tftp"

#include <autoconf.h>
#if defined(CONFIG_ETHERNET_DEBUG)
#define DEBUG
#endif
#include <hermit.h>
#include <io.h>
#include <net/eth.h>
#include <net/udp.h>
#include <net/tftp.h>
#include <string.h>
#include <arch/byteorder.h>

static int create_read_packet(const char *filename, const char *mode,
			      unsigned char *pbuf, unsigned int pbuflen,
			      const char *blksize)
{
	int ptr = 0;
	int filename_len = strlen(filename);
	int mode_len = strlen(mode);
	unsigned short opcode = htons(TFTP_OP_READ);

	if (pbuflen < (2 + filename_len + 1 + mode_len + 1))
		return -1;

	memcpy(&pbuf[ptr], &opcode, 2);
	ptr += 2;

	memcpy(&pbuf[ptr], filename, filename_len);
	ptr += filename_len;

	pbuf[ptr] = '\0';
	ptr += 1;

	memcpy(&pbuf[ptr], mode, mode_len);
	ptr += mode_len;

	pbuf[ptr] = '\0';
	ptr += 1;

	if (blksize && blksize[0]) {
		int blksize_len = strlen(blksize);
		memcpy(&pbuf[ptr], "blksize", 7);
		ptr += 7;

		pbuf[ptr] = '\0';
		ptr += 1;

		memcpy(&pbuf[ptr], blksize, blksize_len);
		ptr += blksize_len;

		pbuf[ptr] = '\0';
		ptr += 1;
	}

	return ptr;
}

static int create_ack_packet(const unsigned short __block,
			     unsigned char *pbuf, unsigned int pbuflen)
{
	int ptr = 0;
	unsigned short opcode = htons(TFTP_OP_ACK);
	unsigned short block = htons(__block);

	if (pbuflen < 4)
		return -1;

	memcpy(&pbuf[ptr], &opcode, 2);
	ptr += 2;

	memcpy(&pbuf[ptr], &block, 2);
	ptr += 2;

	return ptr;
}

int tftp_get(const unsigned char *server_ipaddr,
	     const char *filename,
	     size_t *__filesize,
	     addr_t __buf, unsigned int buflen, const char *blksize)
{
	int ret;
	int flag = 0;
	unsigned char send_pbuf[1024];
	int send_pbuflen;
	unsigned char *recv_pbuf;
	unsigned int recv_pbuflen;
	unsigned short port;

	unsigned short opecode;
	unsigned short block;
	unsigned char *rcv_data;
	unsigned int rcv_data_len;
	unsigned char *filebuf = (unsigned char *)__buf;
	unsigned int filesize = 0;
	int progress = 0;
	int req_block = 1;
	int octets;
	int wait_oack = 0;

	if (blksize && blksize[0]) {
		octets = strtol(blksize, NULL, 0);
		wait_oack = 1;
	} else {
		octets = FIXED_BLOCK_SIZE;
	}

	dev_info_r("Filename : %s\n", filename);
	send_pbuflen = create_read_packet(filename, MODE_BIN, send_pbuf,
					  1024, blksize);
	if (send_pbuflen == -1)
		return -1;

	ret = udp_send(server_ipaddr, TFTP_PORT_NO, 30000,
		       send_pbuf, send_pbuflen);
	if (ret == -1)
		return -1;

	if (wait_oack) {
		ret = udp_proc(server_ipaddr, 30000, &port,
			       (void *)&recv_pbuf, &recv_pbuflen);
		if (ret == -1)
			return -1;

		opecode = ntohs(*(unsigned short *)(recv_pbuf + 0));
		switch (opecode) {
		case TFTP_OP_OPTION:
			send_pbuflen = create_ack_packet(0, send_pbuf, 1024);
			if (send_pbuflen == -1)
				return -1;

			ret = udp_send(server_ipaddr, port, 30000,
				       send_pbuf, send_pbuflen);
			if (ret == -1)
				return -1;
			break;

		case TFTP_OP_ERROR:
		{
			unsigned short errcode;
			unsigned char *errmsg;
			errcode = ntohs(*(unsigned short *)(recv_pbuf + 2));
			errmsg = recv_pbuf + 4;
			dev_err("[%d]: %s\n", errcode, errmsg);
			return -2;
		}
		default:
			dev_err("Could not receive OACK.\n");
			return -2;
		}
	}

	do {
		if ((progress++ % 10) == 0)
			dev_info_r("."); /* progress */

		ret = udp_proc(server_ipaddr, 30000, &port,
			       (void *)&recv_pbuf, &recv_pbuflen);
		if (ret == -1)
			return -1;

		opecode = ntohs(*(unsigned short *)(recv_pbuf + 0));
		switch (opecode) {
		case TFTP_OP_DATA:
			block = ntohs(*(unsigned short *)(recv_pbuf + 2));

			if (req_block != block) {
				dev_dbg("receive unknown block: %d (request: %d)\n",
					req_block, block);
				continue;
			} else {
				req_block++;
			}

			rcv_data_len = recv_pbuflen - 4;
			rcv_data = recv_pbuf + 4;

			/* send ACK */
			send_pbuflen = create_ack_packet(block, send_pbuf, 1024);
			if (send_pbuflen == -1)
				return -1;

			ret = udp_send(server_ipaddr, port, 30000,
				       send_pbuf, send_pbuflen);

			if (rcv_data_len < octets)
				flag = 1;

			if (rcv_data_len == 0)
				break;

			if (buflen - filesize < rcv_data_len) {
				dev_err("Out of memory\n");
				return -2;
			}

			memcpy(&filebuf[filesize], rcv_data, rcv_data_len);

			filesize += rcv_data_len;
			break;

		case TFTP_OP_ERROR:
		{
			unsigned short errcode;
			unsigned char *errmsg;
			errcode = ntohs(*(unsigned short *)(recv_pbuf + 2));
			errmsg = recv_pbuf + 4;
			dev_err("[%d]: %s\n", errcode, errmsg);
			return -2;
		}

		default:
			break;
		}
	} while (flag == 0);

	dev_info_r("\n");
	dev_info_r("Filesize : %d\n", filesize);
	dev_info_r("\n");
	*__filesize = filesize;

	return 0;
}
